"""Represent a mix of tracks and/or videos in the reckoning of TIDAL API."""

import json
import logging
import shutil
import sys
from dataclasses import dataclass
from pathlib import Path
from types import SimpleNamespace
from typing import Dict, List, Optional, Set, Tuple, Union
from uuid import uuid4

from requests import HTTPError, Session

from .media import AudioFormat
from .models import (
    TracksEndpointResponseJSON,
    VideosEndpointResponseJSON,
)
from .track import Track
from .utils import TIDAL_API_URL, replace_illegal_characters
from .video import Video

logger = logging.getLogger("__name__")

# union type for type hinting
MixItem = Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]


@dataclass
class Mix:
    mix_id: str
    transparent: bool = False

    def __post_init__(self):
        self.mix_dir: Optional[Path] = None
        self.mix_cover_saved: bool = False

    def set_metadata(self, session: Session):
        """Request from TIDAL API /mixes endpoint"""
        self.metadata: Optional[SimpleNamespace] = request_mix(
            session=session, mix_id=self.mix_id, transparent=self.transparent
        )

        if self.metadata is None:
            return

        self.name = replace_illegal_characters(self.metadata.title)

    def set_items(self, session: Session):
        """Uses data from TIDAL API /mixes/items endpoint to
        populate self.items"""
        try:
            mix_items: Optional[MixesItemsResponseJSON] = retrieve_mix_items(
                session=session, mix_id=self.mix_id, transparent=self.transparent
            )
        except TidalMixError as tme:
            logger.exception(tme.args[0])
            self.items = tuple()
        else:
            if mix_items is None:
                self.items = tuple()
            else:
                self.items: Tuple[Optional[MixItem]] = tuple(
                    filter(None, mix_items.items)
                )

    def set_mix_dir(self, out_dir: Path):
        """Populates self.mix_dir based on self.name, self.mix_id"""
        mix_substring: str = f"{self.name} [{self.mix_id}]"
        self.mix_dir: Path = out_dir / "Mixes" / mix_substring
        self.mix_dir.mkdir(parents=True, exist_ok=True)

    def save_cover_image(self, session: Session, out_dir: Path):
        """Requests self.metadata.image and attempts to write it to disk"""
        if self.mix_dir is None:
            self.set_mix_dir(out_dir=out_dir)
        self.cover_path: Path = self.mix_dir / "cover.jpg"

        if not self.cover_path.exists():
            with session.get(
                url=self.metadata.image, params={k: None for k in session.params}
            ) as r:
                (self.mix_dir / "cover.jpg").write_bytes(r.content)

            self.mix_cover_saved = True
        else:
            if self.cover_path.stat().st_size > 0:
                self.mix_cover_saved = True

    def get_items(
        self, session: Session, audio_format: AudioFormat, no_extra_files: bool
    ) -> Tuple[Optional[Union[Track, Video]]]:
        """Using either Track.get() or Video.get(), attempt to request
        the data for each track or video in self.items."""
        if len(self.items) == 0:
            self.files = {}
            return

        tracks_videos: List[Optional[Union[Track, Video]]] = [None] * len(self.items)
        for i, item in enumerate(self.items):
            if item is None:
                tracks_videos[i] = None
                continue
            elif isinstance(item, TracksEndpointResponseJSON):
                track: Track = Track(track_id=item.id, transparent=self.transparent)
                track.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=self.mix_dir,
                    metadata=item,
                    no_extra_files=no_extra_files,
                    origin_jpg=False,
                )
                tracks_videos[i] = track
            elif isinstance(item, VideosEndpointResponseJSON):
                video: Video = Video(video_id=item.id, transparent=self.transparent)
                video.get(
                    session=session,
                    out_dir=self.mix_dir,
                    metadata=item,
                )
                tracks_videos[i] = video
            else:
                tracks_videos[i] = None
                continue
        else:
            self.tracks_videos: Tuple[Optional[Union[Track, Video]]] = tuple(
                tracks_videos
            )
        return tracks_videos

    def flatten_mix_dir(self):
        """When self.get_items() is called, the tracks and/or videos in
        self.items are downloaded using their self-contained .get() logic;
        this means that they will be downloaded to albums. This function
        "flattens" self.mix_dir, meaning that it moves all downloaded
        audio and video files to self.mix_dir, and removes the various
        subdirectories created"""
        files: List[Dict[int, Optional[str]]] = [None] * len(self.tracks_videos)
        if len(self.tracks_videos) == 0:
            return

        subdirs: Set[Path] = set()
        for i, tv in enumerate(self.tracks_videos, 1):
            if getattr(tv, "outfile") is None:
                try:
                    getattr(tv, "album_dir")
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
                files[i - 1] = {i: None}
                continue

            _path: Optional[Path] = Path(tv.outfile) if tv is not None else None
            # if the item never got turned into a track or video
            if _path is None:
                files[i - 1] = {i: None}
                continue

            # if the track or video didn't download
            if _path.exists():
                if _path.stat().st_size == 0:
                    files[i - 1] = {i: None}
                    continue
            else:
                files[i - 1] = {i: None}
                continue

            # otherwise, move files and clean up
            if isinstance(tv, Track):
                new_path: Path = self.mix_dir / f"{i:03d} - {tv.trackname}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
            elif isinstance(tv, Video):
                new_path: Path = self.mix_dir / f"{i:03d} - {_path.name}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
        else:
            self.files: List[Dict[int, Optional[str]]] = files

        # Find all subdirectories written to
        subdirs: Set[Path] = set()
        for tv in self.tracks_videos:
            if isinstance(tv, Track):
                try:
                    getattr(tv, "album_dir")
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
            elif isinstance(tv, Video):
                subdirs.add(tv.artist_dir)

        # Copy all artist images, artist bio JSON files out
        # of subdirs
        artist_images: Set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*.jpg"):
                if p.name == "cover.jpg":
                    continue
                artist_images.add(p)
        else:
            for artist_image_path in artist_images:
                if artist_image_path.exists():
                    shutil.copyfile(
                        artist_image_path.absolute(),
                        self.mix_dir / artist_image_path.name,
                    )

        artist_bios: Set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*bio.json"):
                artist_bios.add(p)
        else:
            for artist_bio_path in artist_bios:
                if artist_bio_path.exists():
                    shutil.copyfile(
                        artist_bio_path.absolute(),
                        self.mix_dir / artist_bio_path.name,
                    )

        # Remove all subdirs
        for subdir in subdirs:
            if subdir.exists():
                shutil.rmtree(subdir)
        else:
            return self.mix_dir

    def dumps(self):
        return json.dumps(self.files)

    def dump(self, fp=sys.stdout):
        json.dump(self.files, fp)

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        no_extra_files: bool,
    ):
        """The main method of this class, executing a number of other methods
        in a row:
          - self.set_metadata()
          - self.set_items()
          - self.set_mix_dir()
          - self.get_items()
          - self.flatten_playlist_dir()
        Then, if no_extra_files is False,
          - self.save_cover_image()
        """
        self.set_metadata(session)
        if self.metadata is None:
            self.files = {}
            return

        self.set_items(session)
        self.set_mix_dir(out_dir)

        if self.get_items(session, audio_format, no_extra_files) is None:
            logger.critical(f"Could not retrieve mix with ID '{self.mix_id}'")
            self.files = {}
            return

        self.flatten_mix_dir()
        logger.info(f"Mix files written to '{self.mix_dir}'")

        if not no_extra_files:
            self.save_cover_image(session, out_dir)

    def get_elements(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        no_extra_files: bool,
    ):
        """The main method of this class, executing a number of other methods
        in a row:
          - self.set_metadata()
          - self.set_items()
        """
        self.set_metadata(session)
        if self.metadata is None:
            self.files = {}
            return

        self.set_items(session)
        if len(self.items) == 0:
            self.files = {}
            return
        else:
            files: List[Optional[str]] = [None] * len(self.items)

        for i, item in enumerate(self.items):
            if item is None:
                files[i] = None
                continue
            elif isinstance(item, TracksEndpointResponseJSON):
                track: Track = Track(track_id=item.id, transparent=self.transparent)
                track_file: Optional[str] = track.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=out_dir,
                    metadata=item,
                    no_extra_files=no_extra_files,
                    origin_jpg=False,
                )
                files[i] = track_file
            elif isinstance(item, VideosEndpointResponseJSON):
                video: Video = Video(video_id=item.id, transparent=self.transparent)
                video_file: Optional[str] = video.get(
                    session=session,
                    out_dir=self.mix_dir,
                    metadata=item,
                )
                files[i] = video_file
            else:
                files[i] = None
                continue
        else:
            self.files: List[Optional[str]] = files


class TidalMixError(Exception):
    pass


def request_mix(
    session: Session, mix_id: str, transparent: bool = False
) -> Optional[SimpleNamespace]:
    """Request from TIDAL API /pages/mix endpoint. If an error occurs from
    session.get(), None is returned. Otherwise, a typing.SimpleNamespace
    object is returned with some metadata to do with the mix: title,
    description, URL to cover image."""
    url: str = f"{TIDAL_API_URL}/pages/mix?mixId={mix_id}"
    kwargs: dict = {"url": url}
    kwargs["headers"] = {"Accept": "application/json"}

    json_name: str = f"pages-mix-{mix_id}_{uuid4().hex}.json"

    logger.info(f"Requesting from TIDAL API: mixes/{mix_id}")
    with session.get(**kwargs) as resp:
        try:
            resp.raise_for_status()
        except HTTPError as he:
            if resp.status_code == 404:
                logger.warning(
                    "404 Client Error: not found for TIDAL API endpoint pages/mix"
                )
            else:
                logger.exception(he)
            return
        else:
            if transparent:
                Path(json_name).write_text(
                    json.dumps(resp.json(), ensure_ascii=True, indent=4, sort_keys=True)
                )

        d: Dict[str, str] = {
            "title": resp.json().get("title"),
            "description": resp.json().get("rows")[0]["modules"][0]["mix"]["subTitle"],
        }
        d["image"] = (
            resp.json()
            .get("rows", [{}])[0]
            .get("modules")[0]["mix"]["images"]["LARGE"]["url"]
        )

        logger.debug(
            f"{resp.status_code} response from TIDAL API to request: pages/mix"
        )
        return SimpleNamespace(**d)


def request_mixes_items(
    session: Session,
    mix_id: str,
    offset: Optional[int] = None,
    transparent: bool = False,
) -> Optional[Dict]:
    """Request from TIDAL API /mixes/items endpoint. If error arises when
    requesting with 'session'.get(), None is returned. Otherwise, the
    dict object returned by requests.Response.json() is returned."""
    url: str = f"{TIDAL_API_URL}/mixes/{mix_id}/items"
    kwargs: dict = {"url": url}
    kwargs["params"] = (
        {"limit": 100} if offset is None else {"limit": 100, "offset": offset}
    )
    kwargs["headers"] = {"Accept": "application/json"}
    json_name: str = f"mixes-{mix_id}-items_{uuid4().hex}.json"

    data: Optional[dict] = None
    logger.info(f"Requesting from TIDAL API: mixes/{mix_id}/items")
    with session.get(**kwargs) as resp:
        try:
            resp.raise_for_status()
        except HTTPError as he:
            if resp.status_code == 404:
                logger.warning(
                    f"404 Client Error: not found for TIDAL API endpoint mixes/{mix_id}/items"
                )
            else:
                logger.exception(he)
        else:
            if transparent:
                Path(json_name).write_text(
                    json.dumps(resp.json(), ensure_ascii=True, indent=4, sort_keys=True)
                )
                data = resp.json()
                logger.debug(
                    f"{resp.status_code} response from TIDAL API to request: mixes/{mix_id}/items"
                )
            else:
                data = resp.json()
                logger.debug(
                    f"{resp.status_code} response from TIDAL API to request: mixes/{mix_id}/items"
                )
        finally:
            return data


@dataclass(frozen=True)
class MixesItemsResponseJSON:
    """The response from the TIDAL API endpoint /mixes/<ID>/items
    is modeled by this class."""

    limit: int
    offset: int
    total_number_of_items: int
    items: Tuple[
        Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]
    ]


def mix_items_response_json_maker(
    mixes_response: Dict[str, Union[int, List[dict]]],
) -> "MixesItemsResponseJSON":
    """This function massages the response from the TIDAL API endpoint
    mixes/items into a format that MixesItemsResponseJSON.from_dict()
    can ingest. Returns a MixesItemsResponseJSON instance"""
    init_args: dict = {}
    init_args["limit"] = mixes_response.get("limit")
    init_args["offset"] = mixes_response.get("offset")
    init_args["total_number_of_items"] = mixes_response.get("totalNumberOfItems")

    items: Tuple[SimpleNamespace] = tuple(
        SimpleNamespace(**d) for d in mixes_response["items"]
    )
    if len(items) == 0:
        return

    mixes_items: List[
        Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]
    ] = [None] * init_args["total_number_of_items"]

    for i, namespace in enumerate(items):
        if namespace.type == "track":
            try:
                mix_item = TracksEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                _msg: str = (
                    f"TidalMixError: unable to parse playlist item [i] "
                    f"with type '{namespace.type}'"
                )
                logger.warning(_msg)
                logger.debug(e)
                # value stays None
            else:
                mixes_items[i] = mix_item
        elif namespace.type == "video":
            try:
                mix_item = VideosEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                _msg: str = (
                    f"TidalMixError: unable to parse mix item [i] "
                    f"with type '{namespace.type}'"
                )
                logger.warning(_msg)
                logger.debug(e)
                # value stays None
            else:
                mixes_items[i] = mix_item
        else:
            continue  # value stays None
    init_args["items"] = tuple(mixes_items)

    return MixesItemsResponseJSON(**init_args)


def retrieve_mix_items(
    session: Session, mix_id: str, transparent: bool = False
) -> Optional["MixesItemsResponseJSON"]:
    """The pattern for mix items retrieval does not follow the
    requesting.request_* functions, hence its implementation here. N.b.
    if the first response from /mixes/<ID>/items endpoint indicates that
    the playlist contains more than 100 items, multiple requests will be
    sent until all N > 100 items are retrieved.
    """
    mixes_items_response_json: Optional["MixesItemsResponseJSON"] = None
    mixes_response: Optional[dict] = request_mixes_items(
        session=session, mix_id=mix_id, transparent=transparent
    )
    if mixes_response is None:
        raise TidalMixError(f"Could not retrieve the items in mix '{mix_id}'")

    total_number_of_items: Optional[int] = mixes_response.get("totalNumberOfItems")
    if total_number_of_items is None:
        raise TidalMixError(
            f"TIDAL API did not respond with number of items in mix '{mix_id}'"
        )
    else:
        logger.info(f"Mix '{mix_id}' is comprised of {total_number_of_items} items")

    num_items: int = len(mixes_response.get("items", []))
    if num_items == 0:
        raise TidalMixError(
            f"TIDAL API did not return any mix items for mix '{mix_id}'"
        )

    if num_items < total_number_of_items:
        items_to_retrieve: int = total_number_of_items
        all_items_mixes_response: dict = mixes_response
        if total_number_of_items > 100:
            items_list: List[dict] = mixes_response.pop("items")
            offset: int = 100
            while items_to_retrieve > 0:
                mr: Optional[dict] = request_mixes_items(
                    session=session, mix_id=mix_id, offset=offset
                )
                if (mr is not None) and ((mr_items := mr.get("items")) is not None):
                    items_list += mr_items
                    offset += 100
                    items_to_retrieve -= 100
                else:
                    logger.exception(
                        TidalMixError(
                            f"Could not retrieve more than {len(items_list)} "
                            f"elements of mix '{mix_id}'. Continuing "
                            "without the remaining "
                            f"{total_number_of_items - len(items_list)}"
                        )
                    )
        else:
            all_items_mixes_response = mixes_response
    else:
        all_items_mixes_response = mixes_response

    try:
        mixes_items_response_json: Optional["MixesItemsResponseJSON"] = (
            mix_items_response_json_maker(mixes_response=all_items_mixes_response)
        )
    except Exception as e:
        logger.exception(TidalMixError(e.args[0]))
    finally:
        return mixes_items_response_json



"""Represent a video in the reckoning of TIDAL API."""

import json
import logging
import sys
import urllib
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional

import ffmpeg
import m3u8
import mutagen
from requests import Session

from .hls import TidalM3U8Error, playlister, variant_streams
from .media import TAG_MAPPING
from .models import (
    VideosContributorsResponseJSON,
    VideosEndpointResponseJSON,
    VideosEndpointStreamResponseJSON,
)
from .requesting import request_video_contributors, request_video_stream, request_videos
from .utils import temporary_file

logger = logging.getLogger("__name__")


class VideoFormat(str, Enum):
    high = "HIGH"
    medium = "MEDIUM"
    low = "LOW"
    audio_only = "AUDIO_ONLY"


@dataclass
class Video:
    video_id: int
    transparent: bool = False

    def __post_init__(self):
        self.tags: dict = {}
        self.codec: str = "mp4"

    def get_metadata(self, session: Session):
        """Request from TIDAL API /videos endpoint and set self.metadata
        attribute as None or VideosEndpointResponseJSON instance. N.b.,
        self.metadata.name is a sanitized version of self.metadata.title"""
        self.metadata: Optional[VideosEndpointResponseJSON] = request_videos(
            session=session, video_id=self.video_id, transparent=self.transparent
        )

    def get_contributors(self, session: Session):
        """Request from TIDAL API /videos/contributors endpoint"""
        self.contributors: Optional[VideosContributorsResponseJSON] = (
            request_video_contributors(
                session=session, video_id=self.video_id, transparent=self.transparent
            )
        )

    def get_stream(self, session: Session, video_format=VideoFormat.high):
        """Populates self.stream by requesting from TIDAL API
        /videos/playbackinfopostpaywall endpoint"""
        self.stream: Optional[VideosEndpointStreamResponseJSON] = request_video_stream(
            session=session,
            video_id=self.video_id,
            video_quality=video_format.value,
            transparent=self.transparent,
        )

    def get_m3u8(self, session: Session):
        """This method sets self.m3u8, an m3u8.M3U8 object
        following the HTTP Live Streaming specification; parsed from
        self.stream. I.e., self.get_stream() needs to have been executed
        before calling this method. N.b. self.m3u8 almost certainly will
        be a multivariant playlist, meaning further processing of its
        contents will be necessary."""
        self.m3u8: m3u8.M3U8 = playlister(session=session, vesrj=self.stream)

    def set_urls(self, session: Session):
        """This method uses self.m3u8, an m3u8.M3U8 object that is variant:
        (https://developer.apple.com/documentation/http-live-streaming/creating-a-multivariant-playlist)
        It retrieves the highest-quality .m3u8 in its .playlists attribute,
        and sets self.urls as the list of strings from that m3u8.M3U8 object"""
        # for now, just get the highest-bandwidth playlist
        m3u8_files: List[str] = variant_streams(self.m3u8, session, return_urls=True)
        m3u8_parse_results = (urllib.parse.urlparse(url=f) for f in m3u8_files)
        if not all(u.netloc == "vmz-ad-cf.video.tidal.com" for u in m3u8_parse_results):
            raise TidalM3U8Error(
                f"HLS media segments are not available for video {self.video_id}"
            )
        self.urls: List[str] = m3u8_files

    def set_artist_dir(self, out_dir: Path):
        """Set self.artist_dir, which is the subdirectory of `out_dir`
        with name `self.metadata.artist.name`"""
        self.artist_dir: Path = out_dir / self.metadata.artist.name.replace(
            "..", ""
        ).replace("/", "and")
        self.artist_dir.mkdir(parents=True, exist_ok=True)

    def set_filename(self, out_dir: Path):
        """Set self.filename, which is constructed from self.metadata.name
        and self.stream.video_quality and self.codec"""
        self.filename: str = f"{self.metadata.name} [{self.metadata.id}].{self.codec}"
        self.filename: str = f"{self.metadata.name} [{self.metadata.id}].{self.codec}"

    def set_outfile(self):
        """Uses self.artist_dir and self.metadata and self.filename
        to craft the pathlib.Path object, self.outfile, that is a
        reference to where the video will be written on disk."""
        self.outfile: Path = self.artist_dir / self.filename
        self.absolute_outfile: str = str(self.outfile.absolute())

        if (self.outfile.exists()) and (self.outfile.stat().st_size > 0):
            logger.info(
                f"Video {self.absolute_outfile} already exists "
                "and therefore will not be overwritten"
            )
            return
        else:
            return self.outfile

    def download(self, session: Session, out_dir: Path) -> Optional[Path]:
        """Requests the HLS video files that constitute self.urls.
        Writes HLS bytes to a temporary file, then uses FFmpeg to write the
        video data to self.outfile"""
        download_params: Dict[str, None] = {k: None for k in session.params}
        # self.outfile should already have been set by self.set_outfile()
        request_headers: Dict[str, str] = (
            {"sessionId": session.session_id}
            if session.session_id is not None
            else {}
        )
        logger.info(
            f"Writing video {self.video_id} to '{str(self.outfile.absolute())}'"
        )

        with temporary_file(suffix=".m2t") as tf:
            for i, u in enumerate(self.urls, 1):
                logger.debug(
                    f"\tRequesting part {i} of video {self.video_id}: {u.split('?')[0]}"
                )
                with session.get(
                    url=u, headers=request_headers, params=download_params
                ) as download_response:
                    if not download_response.ok:
                        logger.warning(f"Could not download {self}")
                        return None
                    tf.write(download_response.content)
            tf.seek(0)
            self.outfile.write_bytes(Path(tf.name).read_bytes())

            if self.outfile.exists() and self.outfile.stat().st_size > 0:
                logger.info(
                    f"Video {self.video_id} written to '{self.absolute_outfile}'"
                )

            try:
                ffmpeg.input(tf.name, hide_banner=None, y=None).output(
                    self.absolute_outfile,
                    vcodec="copy",
                    acodec="copy",
                    loglevel="quiet",
                    **{"movflags": "+faststart"},
                ).run()
            except ffmpeg.Error:
                logger.warning(
                    f"Could not convert video {self.video_id} with FFmpeg: "
                    "metadata will not be added and format will stay as MPEG-TS"
                )

        return self.outfile

    def craft_tags(self):
        """Using the TAG_MAPPING dictionary, write the correct values of
        various metadata tags to the file. Videos are AVC1 video, AAC audio in
        MP4 container, so see Kodi reference:
        https://kodi.wiki/view/Video_file_tagging#Overview_and_Comparison"""
        tags = dict()
        tag_map = {k: v["m4a"] for k, v in TAG_MAPPING.items()}

        logger.info(
            f"Adding metadata tags to video {self.video_id} using "
            f"data returned from videos/{self.video_id} API endpoint"
        )
        tags[tag_map["artist"]] = ";".join((a.name for a in self.metadata.artists))
        tags[tag_map["artists"]] = [a.name for a in self.metadata.artists]
        tags[tag_map["date"]] = str(self.metadata.release_date.date())
        tags["hdvd"] = (2,)  # 1080p/i video quality
        tags[tag_map["location"]] = self.contributors.location
        tags[tag_map["media"]] = ["Digital Media"]
        tags["pcst"] = 0  # I.e. False because we are not working with podcasts here
        tags["stik"] = (6,)  # https://exiftool.org/TagNames/QuickTime.html
        tags[tag_map["title"]] = self.metadata.title
        tags["\xa9url"] = f"https://tidal.com/browse/video/{self.video_id}"

        logger.info(
            f"Adding metadata tags to video {self.video_id} using data "
            f"returned from videos/{self.video_id}/contributors API endpoint"
        )
        # Composer
        logger.debug("Adding tag for video composer, if the contributor(s) exist(s)")
        try:
            _credits_tag: str = ";".join(self.contributors.composer)
        except (TypeError, AttributeError):  # NoneType problems
            pass
        else:
            tags[tag_map["composer"]] = _credits_tag

        # Director
        logger.debug("Adding tag for video director, if the contributor(s) exist(s)")
        for tag in {"director", "film_director", "video_director"}:
            try:
                _credits_tag: str = ";".join(getattr(self.contributors, tag))
            except (TypeError, AttributeError):  # NoneType problems
                continue
            else:
                tags[tag_map["director"]] = _credits_tag

        # Engineer
        logger.debug("Adding tag for video engineer, if the contributor(s) exist(s)")
        for tag in {"engineer", "mastering_engineer", "vocal_engineer"}:
            try:
                # self.contributors.mastering_engineer is of type Tuple[str]
                _credits_tag: str = ";".join(getattr(self.contributors, tag))
            except (TypeError, AttributeError):  # NoneType problems
                pass
            else:
                tags[tag_map["engineer"]] = _credits_tag

        # Location
        logger.debug("Adding tag for video location, if the contributor(s) exist(s)")
        try:
            _credits_tag: str = ";".join(self.contributors.location)
        except (TypeError, AttributeError):  # NoneType problems
            pass
        else:
            tags[tag_map["location"]] = _credits_tag

        # Lyricist
        logger.debug("Adding tag for video lyricist, if the contributor(s) exist(s)")
        try:
            # self.contributors.lyricist is of type Optional[Tuple[str]]
            _credits_tag: str = ";".join(self.contributors.lyricist)
        except (TypeError, AttributeError):  # NoneType problems
            pass
        else:
            tags[tag_map["lyricist"]] = _credits_tag

        # Mixing
        logger.debug("Adding tag for video mixer, if the contributor(s) exist(s)")
        for tag in ("assistant_mixer", "mixer", "mixing_engineer"):
            try:
                _credits_tag: str = ";".join(getattr(self.contributors, tag))
            except (TypeError, AttributeError):  # NoneType problems
                continue
            else:
                tags[tag_map["mixer"]] = _credits_tag

        # Performer
        logger.debug("Adding tag for video performer, if the contributor(s) exist(s)")
        try:
            _credits_tag: str = ";".join(self.contributors.associated_performer)
        except (TypeError, AttributeError):  # NoneType problems
            pass
        else:
            tags[tag_map["performer"]] = _credits_tag

        # Producer
        logger.debug("Adding tag for video producer, if the contributor(s) exist(s)")
        for tag in {"film_producer", "producer", "video_producer"}:
            try:
                _credits_tag: str = ";".join(getattr(self.contributors, tag))
            except (TypeError, AttributeError):  # NoneType problems
                continue
            else:
                tags[tag_map["producer"]] = _credits_tag

        # Publisher
        logger.debug("Adding tag for video publisher, if the contributor(s) exist(s)")
        try:
            # self.contributors.music_publisher is Optional[Tuple[str]]
            _credits_tag: str = ";".join(self.contributors.music_publisher)
        except (TypeError, AttributeError):  # NoneType problems
            pass
        else:
            tags[tag_map["publisher"]] = _credits_tag

        # Have to convert to bytes the values of the tags starting with '----'
        for k, v in tags.copy().items():
            if k.startswith("----"):
                if isinstance(v, str):
                    tags[k] = v.encode("UTF-8")
                elif isinstance(v, list):
                    tags[k] = [s.encode("UTF-8") for s in v]

        self.tags: dict = {k: v for k, v in tags.items() if v is not None}

    def set_tags(self):
        """Instantiate a mutagen.File instance, add self.tags to it, and
        save it to disk. If video container is still in MPEG-TS format,
        this is expected to fail"""
        try:
            self.mutagen = mutagen.File(self.outfile)
        except Exception:
            logger.warning(f"Unable to write metadata tags to {self.video_id}")
            return

        self.mutagen.clear()
        self.mutagen.update(**self.tags)
        self.mutagen.save()

    def get(
        self,
        session: Session,
        out_dir: Path,
        metadata: Optional["VideosEndpointResponseJSON"] = None,
    ) -> Optional[str]:
        """The main method of this class. Executes a number of other methods
        in a row:
          - self.get_metadata()
          - self.get_contributors()
          - self.get_stream()
          - self.get_m3u8()
          - self.set_urls()
          - self.set_artist_dir()
          - self.set_filename()
          - self.set_outfile()
          - self.download()
          - self.craft_tags()
          - self.set_tags()
        """
        if metadata is None:
            self.get_metadata(session)
        else:
            self.metadata = metadata

        if self.metadata is None:
            return None

        self.get_contributors(session)
        self.get_stream(session)
        if self.stream is None:
            return None
        self.get_m3u8(session)
        self.set_urls(session)
        self.set_artist_dir(out_dir)
        self.set_filename(out_dir)
        outfile: Optional[Path] = self.set_outfile()
        if outfile is None:
            return None

        if self.download(session, out_dir) is None:
            return None

        self.craft_tags()
        self.set_tags()

        return self.absolute_outfile

    def dump(self, fp=sys.stdout):
        """This method emulates stdlib json.dump(). In particular,
        it sends to 'fp' the JSON-formatted dict
        {self.metadata.title: self.absolute_outfile}"""
        json.dump({self.metadata.title: self.absolute_outfile}, fp)

    def dumps(self) -> str:
        """This method emulates stdlib json.dumps(). In particular,
        it returns the JSON-formatted str from the dict
        {self.metadata.title: self.absolute_outfile}"""
        return json.dumps({self.metadata.title: self.absolute_outfile})
